/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <cmath>
#include <iostream>
 
#include "sGFX_pieces.hpp"

#include "sGFX_colorchooser.hpp"
#include "sGFX_cube.hpp"

//**************************************************************************

SomaPiecesGFX::SomaPiecesGFX(const SomaProb *const p,const SomaVector<double>& scolor_in):
                             pieces(p->get_num_pieces()),
                             cubes(p->get_num_pieces()),
                             rotations(p->get_num_pieces()),
                             scolor(scolor_in){
                               
  SomaColorChooser rgb(pieces);
  
  for(unsigned int i=0;i<pieces;i++){
    
    cubes[i] = new SomaCubeArray(p->get_piece(i)->get_num_blocks());
  
    SomaVector<int> b;
    const SomaPiece *const pc = p->get_piece(i);
    SomaVector<double> offset = pc->get_offset();
    
    for(unsigned int j=0;j<pc->get_num_blocks();j++){
      pc->get(j,&b);
      SomaVector<double> pos(b);
      cubes[i]->insert(pos+offset,rgb.get(i));
    }
  }
  
  distance = -8;
}
 
//**************************************************************************

SomaPiecesGFX::~SomaPiecesGFX(){
  for(unsigned int i=0;i<pieces;i++){
    delete cubes[i];
  }
}
  
//**************************************************************************

void SomaPiecesGFX::resize(const unsigned int width_in,
                           const unsigned int height_in){
  
  const double k = sqrt( width_in * height_in / double(pieces) );
  if(k==0) return;
  
  width = width_in;
  height = height_in;
  
  widthdiv = (unsigned int) ceil( width/k );
  heightdiv = (unsigned int) ceil( height/k );
  
  while(widthdiv * (heightdiv-1) >= pieces) heightdiv--;
  while((widthdiv-1) * heightdiv >= pieces) widthdiv--;
  
  smallwidth = width/widthdiv;
  smallheight = height/heightdiv;
  
  widthmargin = (width - smallwidth*widthdiv)/2;
  heightmargin = (height - smallheight*heightdiv)/2;
  
  w = smallwidth;
  h = smallheight;
  
  if(w<h){
    h = h/w*0.2;
    w = 0.2;
  }else{
    w=w/h*0.2;
    h=0.2;
  }
   
}
  
//**************************************************************************

void SomaPiecesGFX::draw(){
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-w, w, -h, h, 0.5, 15.5); 
  glViewport(0,0,width,height);
  
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  for(unsigned int i=0;i<pieces;i++){
      
    const unsigned int y = i / widthdiv;
    const unsigned int x = i % widthdiv;

    glViewport(widthmargin+x*smallwidth,height-(heightmargin+(y+1)*smallheight),smallwidth,smallheight);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
  
    glTranslated(0,0,distance);
    glMultMatrixd(rotations[i].mtr);
    
    for(unsigned j=0;j<cubes[i]->size();j++){
      cubes[i]->get(j)->draw(GL_RENDER);
    }
  }
  
  glViewport(0,0,width,height); 
  glClear( GL_DEPTH_BUFFER_BIT);
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0,width,0,height,0,1);
 
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  bool light_enabled = glIsEnabled(GL_LIGHTING);
  bool material_enabled = glIsEnabled(GL_COLOR_MATERIAL);
  
  glDisable(GL_LIGHTING);
  glDisable(GL_COLOR_MATERIAL);
  
  glColor3d(scolor.x,scolor.y,scolor.z);
  
  glBegin(GL_QUADS);
    glVertex3d(          -2,     -2, 0);
    glVertex3d( widthmargin,     -2, 0);
    glVertex3d( widthmargin, height, 0);
    glVertex3d(          -2, height, 0);
    
    glVertex3d( widthmargin+widthdiv*smallwidth,       -1, 0);
    glVertex3d( width+1,                               -1, 0);
    glVertex3d( width+1,                         height+1, 0);
    glVertex3d( widthmargin+widthdiv*smallwidth, height+1, 0);
    
    glVertex3d(      -2,           -2, 0);
    glVertex3d( width+1,           -2, 0);
    glVertex3d( width+1, heightmargin+1, 0);
    glVertex3d(      -2, heightmargin+1, 0);
    
    glVertex3d(      -2,                           height+1, 0);
    glVertex3d(      -2, heightmargin+heightdiv*smallheight, 0);
    glVertex3d( width+1, heightmargin+heightdiv*smallheight, 0);
    glVertex3d( width+1,                           height+1, 0);
  glEnd();
  
  glBegin(GL_LINES);

  for(unsigned int x=0;x<=widthdiv;x++){
    glVertex3d(widthmargin+x*smallwidth,     0,0);
    glVertex3d(widthmargin+x*smallwidth,height,0);
  }
  if(widthmargin==0){
    glVertex3d(width-1,      0,0);
    glVertex3d(width-1, height,0);
  }
  

  for(unsigned int y=0;y<=heightdiv;y++){
    glVertex3d(     0,height-(heightmargin+y*smallheight), 0);
    glVertex3d( width,height-(heightmargin+y*smallheight), 0);
  }
  if(heightmargin==0){
    glVertex3d(     0, 0, 0);
    glVertex3d( width, 0, 0);
  }

  glEnd();
  
  if(light_enabled) glEnable(GL_LIGHTING);
  if(material_enabled) glEnable(GL_COLOR_MATERIAL);
}
 
//**************************************************************************

void SomaPiecesGFX::mouseClick(const unsigned int x,const unsigned int y,
                               const MOUSEBUTTON button){

    selected = x/smallwidth+widthdiv*(y/smallheight);
}
 
//**************************************************************************

void SomaPiecesGFX::mouseMove(const int deltax,const int deltay,
                              const MOUSEBUTTON button){
  if(button == MIDDLE){
    distance += 0.2 * deltay; 
    if(distance < -11) distance = -11;
    if(distance > -5) distance = -5;   
  }else if(button == LEFT && selected < pieces){
    rotations[selected].rotate(SomaMatrix::Y,0.015*deltax);
    rotations[selected].rotate(SomaMatrix::X,0.015*deltay);
  }else if(button == RIGHT){
    for(unsigned int i=0;i<pieces;i++){
      rotations[i].rotate(SomaMatrix::Y,0.015*deltax);
      rotations[i].rotate(SomaMatrix::X,0.015*deltay);   
    }
  }
}
 
//**************************************************************************

