/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef soma_problem
#define soma_problem

#include <vector>
#include <list>
#include <iterator>

#include "s_io.hpp"
#include "s_vector.hpp"
#include "s_piece.hpp"
#include "s_solution.hpp"

//**************************************************************************

class SomaProb : public SomaIO{
private:
  std::vector<SomaPiece>* pieces;
  unsigned int num_pieces;
  
  SomaSolutions *solution;
  
  std::list<SomaVector<int> >* blocks;
  unsigned int max_blocks;
  unsigned int num_blocks;
  
  mutable std::list<SomaVector<int> >::const_iterator iter;
  
  SomaProb();
  static bool check_solution(const unsigned int num_pieces,
                             const std::vector<SomaPiece> *const pieces,
                             const unsigned int num_blocks,
                             const std::list<SomaVector<int> > *const blocks,
                             const SomaSolution *solution);
public:
  
  /** Create a new problem object from a stream.
  The stream has to contain valid input of course.
  The object returned must be deleted later. */ 
  static SomaProb* create_problem(std::istream& strm);
  
  /** Delete this object. */
  ~SomaProb();
  
  /** Get the number of pieces in this problem. */
  inline unsigned int get_num_pieces()const{return num_pieces;}

  /** Get a pointer to a specified piece. */  
  const SomaPiece* get_piece(const unsigned int piece_num)const;
  
  /** Set a solution. */
  void add_solution(SomaSolution& solution_in);
  
  /** Get a pointer to solution data. */
  const SomaSolutions* get_solution();
  
  /** Get the number of blocks currently inserted. */
  inline unsigned int get_num_blocks()const{return num_blocks;}
  
  /** Get the number of blocks which maximally could be inserted. */
  inline unsigned int get_max_blocks()const{return max_blocks;}
  
  /** Reset the internal list of blocks.
  Returns true if there are any blocks in the list.*/
  bool reset_blocks()const;
  
  /** Get the current block from the list and increase the internal 
  list pointer. Returns true if there are any more blocks to get from 
  the list. */
  bool get_block(SomaVector<int> *const xyz)const;

  /** Returns true if there are any more blocks to get from the list. */ 
  bool more_blocks()const;
  
  /** Inserts the new given block if max_blocks has not been reached
  and this block is not already inserted. Returns true if
  insertion was successfull, false otherwise. */
  bool insert_block(const SomaVector<int>& xyz);

  /** Removes the given block if it is in this object. Returns true if
  removal was successful, false otherwise. */
  bool remove_block(const SomaVector<int>& xyz);

  /** Remove all blocks. */
  void clear_blocks();
  
  /** Check if this block is in the problem. */
  bool is_block(const SomaVector<int>& xyz)const;
  
  /** Centers the problem on integer coordinates around (0,0,0). Returns
      the offset that needs to be added to each block's coordinate to
      center the problem completly.*/
  SomaVector<double> center();

  /** Check if problem is currently solved. */
  inline bool is_solved(){return (solution != 0);}
  
  /** Reset the object after a successfull solve(),
  so the blocks can be edited again. */
  void unsolve();
  
  /** Write object values to a stream. This includes
  the pieces and problem blocks and also the solution. */
  void write(std::ostream& strm = std::cout)const;
  
  /** Read object values from a stream. The input has to cointain pieces
  and may contain problem blocks and a solution. Returns true if
  reading was successful, in that case the object values have
  changed. Otherwise false will be returned and the object values
  will remain. */
  bool read(std::istream& strm = std::cin);
};

//**************************************************************************

#endif
