/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cstdlib>
#include <ctime>

#include <limits.h>
 
#include <list>
#include <iterator> 
 
#include "s_solve.hpp" 
#include "s_vector.hpp"
#include "s_piece.hpp"
#include "s_solution.hpp"

//**************************************************************************

bitvector SomaSolve::flags[8*sizeof(bitvector)+1];

//**************************************************************************

void SomaSolve::initFlags(){
  flags[0] = 0;
  flags[1] = 1;
  for(unsigned int i=1;i<8*sizeof(bitvector);i++){
    flags[i+1] = 2*flags[i];
  }
}

//**************************************************************************

void SomaSolve::printT(const bitvector val){
  for(unsigned int i=8*sizeof(bitvector);i>0;i--){
    if(val & flags[i]) std::cout << "#";
    else std::cout << "_";
    if((i-1)%8 == 0 && i!=1) std::cout << " ";
  }
  std::cout << std::endl;
}
  
//**************************************************************************

unsigned int SomaSolve::calc_blockmax(const SomaProb *const p){
  // how big is the largest piece ?
  // this will determine the size of our 3d space to prevent array overflows
  unsigned int blockmax = p->get_piece(0)->get_num_blocks();
  for(unsigned int i=0;i<p->get_num_pieces();i++){
    if(p->get_piece(i)->get_num_blocks() > blockmax) 
      blockmax =  p->get_piece(i)->get_num_blocks();
  }
  // we need at least a "safety border" of one block to prevent array overflows 
  // in the connectCheck, otherwise we may decrease it by one
  if(blockmax > 1) blockmax--;
  
  return blockmax;
}
  
//**************************************************************************

SomaVector<int> SomaSolve::calc_min(const SomaProb *const p){
  SomaVector<int> block,min;
  if(p->reset_blocks()){
    p->get_block(&block);
    min.x = block.x;
    min.y = block.y;
    min.z = block.z;
  }
  while(p->more_blocks()){
    p->get_block(&block);
    
    if(block.x < min.x) min.x = block.x;
    if(block.y < min.y) min.y = block.y;
    if(block.z < min.z) min.z = block.z;
  }

  return min;  
}
  
//**************************************************************************

SomaVector<int> SomaSolve::calc_max(const SomaProb *const p,const SomaVector<int>& min,const unsigned int blockmax){
  SomaVector<int> block,max;
  if(p->reset_blocks()){
    p->get_block(&block);
    max.x = block.x;
    max.y = block.y;
    max.z = block.z;
  }
  while(p->more_blocks()){
    p->get_block(&block);
    
    if(block.x > max.x) max.x = block.x;
    if(block.y > max.y) max.y = block.y;
    if(block.z > max.z) max.z = block.z;
  }
  
  max.x = max.x - min.x + 2*blockmax + 1;
  max.y = max.y - min.y + 2*blockmax + 1;
  max.z = max.z - min.z + 2*blockmax + 1;

  return max;  
}
  
//**************************************************************************

SomaSolve* SomaSolve::createSolver(SomaProb *const p){
 if(p->get_num_blocks() <= 8*sizeof(bitvector))
   return new SomaSolve(p);
 else 
   return 0;
}
  
//**************************************************************************
 
SomaSolve::SomaSolve(SomaProb *const p_in):p(p_in),
                                           num_blocks(p->get_num_blocks()),
                                           num_pieces(p->get_num_pieces()),
                                           blockmax(calc_blockmax(p)),
                                           min(calc_min(p)),
                                           max(calc_max(p,min,blockmax)),
                                           xmult(max.y*max.z),
                                           ymult(max.z),
                                           val2(xmult << 1),
                                           val3(xmult+ymult),
                                           val4(ymult << 1),
                                           val5(ymult+1){
 
  initFlags();
  
  // connectCheck Stuff
  queue = new unsigned int[num_blocks];
  confound = new unsigned int[num_blocks+1];
  reverse_lookup = new int[num_blocks+1];
  confound[0] = false; // 0 blocks are always found already
  
  knapsack_solvable = new unsigned int[num_pieces*num_blocks];
  for(unsigned int i=0;i<num_pieces;i++){
    for(unsigned int j=0;j<num_blocks;j++){
      knapsack_solvable[i*num_blocks+j] = 0;
    }
  }
  
  // coverCheck Stuff
  cover = new bitvector[num_pieces];
  
  //create array out of blocklist
  space = new unsigned int[max.x*max.y*max.z];

  for(int i=0;i<max.x;i++){
    for(int j=0;j<max.y;j++){
      for(int k=0;k<max.z;k++){
        space[i*xmult + j*ymult + k] = 0;
      }
    }
  }
  
  
  //number blocks 1 ... num_blocks
  SomaVector<int> block;
  unsigned int count = 1;
  p->reset_blocks();
  while(p->more_blocks()){
    p->get_block(&block);
    space[(block.x-min.x+blockmax) * xmult 
        + (block.y-min.y+blockmax) * ymult 
        + (block.z-min.z+blockmax)] = count;
        
    reverse_lookup[count] = (block.x-min.x+blockmax) * xmult 
                          + (block.y-min.y+blockmax) * ymult 
                          + (block.z-min.z+blockmax);
    count++;
  }
 
  createLists();
  
  // stuff for main iteration  
  inuse = new unsigned int[num_pieces];
  up = new unsigned int[num_pieces];
  down = new unsigned int[num_pieces];
  for(unsigned int i=0;i<num_pieces;i++){
    up[i] = down[i] = 0;
  }
  
  inserted = solved = 0;
  for(unsigned int i=1;i<=num_blocks;i++){
    solved |= flags[i];
  }
  
  //to determine which pieces are possible solve knapsack problem
  unsigned int* a = new unsigned int[num_pieces];
  for(unsigned int i=0;i<num_pieces;i++){
    a[i] = p->get_piece(sorted_index[i])->get_num_blocks();
  }
  K = new SomaKnapsack(a,num_blocks,num_pieces);
  delete[] a;
}

//**************************************************************************

SomaSolve::~SomaSolve(){

  delete[] space;
  
  for(unsigned int i=0;i<num_pieces;i++){
     for(unsigned int j=0;j<ial[i];j++){
      delete insarr[i][j];
    }
    delete[] insarr2[i];
    delete[] insarr[i];
  }
  delete[] insarr;
  delete[] insarr2;
  
  delete[] ial;
  
  delete[] inuse;
  delete[] up;
  delete[] down;
  delete K;
  
  // connectCheck Stuff
  delete[] confound;
  delete[] queue;
  delete[] reverse_lookup;
  delete[] knapsack_solvable;
  
  // coverCheck Stuff
  delete[] cover;
  
  delete[] sorted_index;
}

//**************************************************************************
 
void SomaSolve::createLists(){
  // lists containing for each piece each position it could be
  // inserted and which blocks it fills then 
  std::list<SomaSet*>** llists = new std::list<SomaSet*>*[num_pieces];
  std::list<SomaSet*>::iterator it,jt;
  
  for(unsigned int i=0;i<num_pieces;i++){
    llists[i] = new std::list<SomaSet*>;
  }
  
  SomaSet* current;
  SomaVector<int> pos;
  SomaPiece piece(1);
  unsigned int temp;
  
  // check for each piece how it fits into the blocks, for each insertion 
  // save which blocks are filled in a list
  for(unsigned int i=0;i<num_pieces;i++){
    piece.set(*p->get_piece(i));
    const unsigned int blocknum = piece.get_num_blocks();
    current = new SomaSet;
    current->st = 0;

    for(unsigned int j=0;j<SomaPiece::ORIENTATIONS;j++){
      piece.set_orientation(j);
      
      const SomaVector<int> max2(max.x-blockmax,max.y-blockmax,max.z-blockmax);
      for(int l=blockmax;l<max2.x;l++){
        for(int m=blockmax;m<max2.y;m++){
          for(int n=blockmax;n<max2.z;n++){
            // try to insert piece at l,m,n
            
            for(unsigned int k=0;k<blocknum;k++){
              piece.get(k,&pos);
              pos.x+=l; pos.y+=m; pos.z+=n;
              
              if(space[pos.x*xmult + pos.y*ymult + pos.z] != 0){
                current->st |= flags[space[pos.x*xmult + pos.y*ymult + pos.z]];
              }else{
                current->st = 0;
                break;
              }
            }
            
            if(current->st){
              current->pos.set(l,m,n);
              current->ori = j;
  
              llists[i]->push_back(current);

              current = new SomaSet;
              current->st = 0;
            }      
          }
        }
      }
    }
    delete current;
  }
 
  //delete double entrys in lists which result from symmetries in the pieces
  for(unsigned int i=0;i<num_pieces;i++){
    it = llists[i]->begin();
    while(it != llists[i]->end()){
      jt = it; 
      jt++;
      while(jt != llists[i]->end()){
        if((*it)->st == (*jt)->st){
          delete *jt;
          jt = llists[i]->erase(jt);
        }else{
          jt++;
        }
      }
      it++;
    }
  }

  //generate sorted index
  sorted_index = new unsigned int[num_pieces];
  for(unsigned int i=0;i<num_pieces;i++){
    sorted_index[i]  = i;
  }
  
  // sort fore piecesize
  for(unsigned int i=1;i<num_pieces;i++){
    temp = sorted_index[i];
    unsigned int j = i;
    while(j>0 && p->get_piece(temp)->get_num_blocks() <  p->get_piece(sorted_index[j-1])->get_num_blocks()){
      sorted_index[j] = sorted_index[j-1];
      j--;
    }
    sorted_index[j] = temp;
  }
  
  // sort for listlength
  for(unsigned int i=1;i<num_pieces;i++){
    temp = sorted_index[i];
    unsigned int j = i;
    while(j>0 && llists[temp]->size() < llists[sorted_index[j-1]]->size()){
      sorted_index[j] = sorted_index[j-1];
      j--;
    }
    sorted_index[j] = temp;
  }
    
  //copy lists -> array, clean up lists 
  ial = new unsigned int[num_pieces];
  insarr = new SomaSet**[num_pieces];
  insarr2 = new bitvector*[num_pieces];
  for(unsigned int i=0;i<num_pieces;i++){
    ial[i] = llists[sorted_index[i]]->size();
    insarr[i] = new SomaSet*[ial[i]];
    insarr2[i] = new bitvector[ial[i]];
    
    it = llists[sorted_index[i]]->begin();
    for(unsigned int j=0;j<ial[i];j++){
      insarr[i][j] = *it;
      it++;
    }
    
    delete llists[sorted_index[i]];    
  } 
  delete[] llists;

  //make copy of insarr to insarr2
  for(unsigned int i=0;i<num_pieces;i++){
    for(unsigned int j=0;j<ial[i];j++){
      insarr2[i][j] = insarr[i][j]->st;
    }
  }
}
 
//**************************************************************************

void SomaSolve::initConnectCheck(){
  unsigned int dim = 0;
  for(unsigned int i=0;i<num_pieces;i++){
    if(possible[i]){
      dim++;
    }
  }
  
  unsigned int* a = new unsigned int[dim];
  unsigned int j = dim-1;
  
  for(unsigned int i=0;i<num_pieces;i++){
    if(possible[i]){
      a[j--] = p->get_piece(sorted_index[i])->get_num_blocks();      
    }
  }
  
  j = num_blocks;
  SomaKnapsack k(a,1,dim);
  for(unsigned int i=0;i<num_pieces;i++){
    if(possible[i]){
      j -= a[--dim];
      if(dim > 0){
        unsigned int l = i+1;
        while(!possible[l]) l++;
        k.set_dim(dim);
        for(unsigned int b=1;b<=j;b++){
          k.set_b(b);
          knapsack_solvable[(l)*num_blocks+b] = k.next_solution() ? 1 : 0;
        }
      }
      
    }
  }

#ifdef SOMA_VERBOSE  
  for(unsigned int i=0;i<num_pieces;i++){
    for(unsigned int b=0;b<num_blocks;b++){
      std::cout << (knapsack_solvable[i*num_blocks+b] ? "#" : "_");
    }
    std::cout << std::endl;
  }
#endif

  delete[] a;
}

//**************************************************************************

unsigned int SomaSolve::connectCheck(){ 
  unsigned int concurr; 
  unsigned int cindex; 
  unsigned int qs=0,qe=0; 

  unsigned int *pi, i;
  
  // mark inserted blocks as found 
  for(i=1;i<=num_blocks;i++){ 
    confound[i] = !(inserted & flags[i]); 
  } 

  // check open problem space for connected components with bfs   
  // search start point 
  for(i=num_blocks;i>0;i--){ 
     
    if(confound[i]){        
      // new component found  
      queue[qs++] = reverse_lookup[i]; 
      confound[i] = false; 
      concurr = 0; 
       
      // simple bfs 
      while(qs>qe){ 
        concurr++; 
         
        // check in six directions
        
        cindex = queue[qe++] + xmult; 
        pi = confound + space[cindex];
        if(*pi){ 
          queue[qs++] = cindex; 
          *pi = false; 
        } 
        
        cindex -= val2;
        pi = confound + space[cindex];
        if(*pi){ 
          queue[qs++] = cindex; 
          *pi = false; 
        } 
         
        cindex += val3;
        pi = confound + space[cindex];
        if(*pi){ 
          queue[qs++] = cindex; 
          *pi = false; 
        } 
         
        cindex -= val4;
        pi = confound + space[cindex];
        if(*pi){ 
          queue[qs++] = cindex; 
          *pi = false; 
        } 
        
        cindex += val5;
        pi = confound + space[cindex];
        if(*pi){ 
          queue[qs++] = cindex; 
          *pi = false; 
        } 
        
        cindex -= 2;
        pi = confound + space[cindex];
        if(*pi){ 
          queue[qs++] = cindex; 
          *pi = false; 
        } 
         
      } 

      // open space can be filled in knapsack relaxation ??
      if(! knapsack_solvable[ta*(num_blocks)+concurr]) return true;
      
    } 
     
  } 
    
  return false; 
}

//**************************************************************************

unsigned int SomaSolve::coverCheck(){
  bitvector cover1 = inserted;
  bitvector * insarr2i;
  unsigned int i=ta;
  do{
      i += up[i];
      insarr2i = insarr2[i];
      for(unsigned int j=0;j<ial[i];j++){
        if(! (insarr2i[j] & inserted)){
          cover1 |= insarr2i[j];
        }
      }
  }while(up[i] != 0);
  
  cover[ta] = cover1;
  for(unsigned int j=0;j<ial[ta];j++){
    if(! (insarr2[ta][j] & inserted)){
      cover1 |= insarr2[ta][j];
    }
  }
  return !(cover1 == solved); 
}

//**************************************************************************

void SomaSolve::solve(unsigned int num_solutions,
                      const unsigned int randomize){
  unsigned int ia;
  
  if(num_solutions == 0) num_solutions = UINT_MAX;
  
  if(randomize){
    srand(time(0));
    for(unsigned int i=0;i<num_pieces;i++){
      for(unsigned j=0;j<ial[i];j++){
        const unsigned int k = rand()%ial[i];
        bitvector swap = insarr2[i][j];
        insarr2[i][j] = insarr2[i][k];
        insarr2[i][k] = swap;
        
        SomaSet* swap1 = insarr[i][j];
        insarr[i][j] = insarr[i][k];
        insarr[i][k] = swap1;
      }
    }
  }
  
  possible = K->next_solution();
  while(possible){    
    initConnectCheck();
    
#ifdef SOMA_VERBOSE
    std::cout << std::endl << "       ";
    for(unsigned int j=0;j<num_pieces;j++){
      if(possible[j]){
        std::cout.width(4);
        std::cout << j;
      }
    }
    std::cout << std::endl << "size  :";
    for(unsigned int j=0;j<num_pieces;j++){
      if(possible[j]){
        std::cout.width(4);
        std::cout << p->get_piece(sorted_index[j])->get_num_blocks();
      }
    }
    std::cout << std::endl << "lenght:";
    for(unsigned int j=0;j<num_pieces;j++){
      if(possible[j]){
        std::cout.width(4);
        std::cout << ial[j];
      }
    }
    std::cout << std::endl << "pos   :";
#endif      
    
    // build up/down vectors
    ta = ia = 0;
    while(!possible[ia]) ia++;
    for(;ia<num_pieces;ia++){
      if(possible[ia]){
        down[ia] = ta;
        up[ia-ta] = ta;
        ta = 0;
      }
      ta++;
    }

    ta = ia = 0;
    while(!possible[ta]) ta++;       
    coverCheck();
    
#ifdef SOMA_VERBOSE
    std::cout.width(4);
    std::cout << "   0" << std::flush;      
#endif
    
    for(;;){
      
      while(ia != ial[ta]){
        
        if(!(insarr2[ta][ia] & inserted) && ((insarr2[ta][ia] | cover[ta]) == solved)){
          inserted |= insarr2[ta][ia];
          inuse[ta] = ia;
          
          if(solved == inserted){
            
#ifdef SOMA_VERBOSE
            std::cout << "\b\b\b\b";
            std::cout.width(4);
            std::cout << ia << std::flush;
#endif
            writeSolution();
            // return if enough solutions found
            if(p->get_solution()->size() == num_solutions) return;

          }else{
            
            if(up[ta] == 0) goto remove;
            ta += up[ta];
            
            if(connectCheck() || coverCheck()){
              ta -= down[ta];
              goto remove;
            }
            
#ifdef SOMA_VERBOSE
            std::cout << "\b\b\b\b";
            std::cout.width(4);
            std::cout << ia;
            std::cout << "   0" << std::flush;      
#endif                             

            ia = 0;  
            continue;
          }
          
          remove:
          inserted ^= insarr2[ta][ia];
        }
        ia++;
        
      }
      
#ifdef SOMA_VERBOSE
      std::cout << "\b\b\b\b";
      std::cout.width(4);
      std::cout << ia;
#endif   
      
      if(down[ta] == 0) break;
      
      ta -= down[ta];
      ia = inuse[ta];
      inserted ^= insarr2[ta][ia];
      ia++;
      
#ifdef SOMA_VERBOSE
      std::cout << "\b\b\b\b\b\b\b\b";
      std::cout.width(4);
      std::cout << ia << std::flush;
#endif
      
    }  
    
    possible = K->next_solution();  
  }

}

//**************************************************************************

void SomaSolve::writeSolution(){
  SomaSolution solution(num_pieces);
  for(unsigned int i=0;i<num_pieces;i++){
    if(possible[i]){
      SomaVector<int> sv = insarr[i][inuse[i]]->pos + min;
      sv.subtract(blockmax,blockmax,blockmax);
      solution.insert(sorted_index[i],sv,insarr[i][inuse[i]]->ori);
    }
  }
  p->add_solution(solution);
}

//**************************************************************************

